"""Traces subsystem"""
from rsfsup.common import Subsystem, validate

_MODES = {"WRIT": "WRITE", "AVER": "AVERAGE", "MAXH": "MAXHOLD"}


class Trace(Subsystem, kind="Trace"):
    """Trace subsystem

    Attributes:
        instr (Fsup)
        num (int): trace number {1, 2, 3}
    """

    def __init__(self, instr, num):
        super().__init__(instr)
        self._num = num
        self._name = f"trace_{num}"

    @property
    def mode(self):
        """value (str): {WRITE, AVERAGE, MAXHOLD}"""
        value = self._visa.query(f"DISP:WIND1:TRAC{self._num}:MODE?")
        return _MODES[value]

    @mode.setter
    @validate
    def mode(self, value):
        self._visa.write(f"DISP:WIND1:TRAC{self._num}:MODE {value}")

    @property
    def state(self):
        """value (str): {ON , OFF}"""
        value = int(self._visa.query(f"DISP:WIND1:TRAC{self._num}:STATE?"))
        return "ON" if value else "OFF"

    @state.setter
    @validate
    def state(self, value):
        self._visa.write(f"DISP:WIND1:TRAC{self._num}:STATE {value}")

    @property
    def y_unit(self):
        """(str) Y axis unit"""
        result = self._visa.query(f"DISP:WIND1:TRAC:Y:UNIT?")
        noise_units = {
            "DBC/HZ": "dBc/Hz",
        }
        return noise_units.get(result, result)

    @property
    def data(self):
        """(tuple): (X, Y) values"""
        # In SSA mode, trace data is both x and y interleaved
        data = self._visa.query_ascii_values(f"TRACE? TRACE{self._num}")
        x = data[::2]
        y = data[1::2]
        return (x, y)



"""Display subsystem specific to SSA mode"""
from rsfsup.common import Subsystem, validate


class Display(Subsystem, kind="Display"):
    """Display subsystem

    Attributes:
        instr (Fsup)
    """

    def __init__(self, instr):
        super().__init__(instr)
        self._psave_holdoff = 0
        self._visa.write(f"DISPLAY:PSAVE:STATE OFF")

    @property
    def pwr_save(self):
        """value (int): screen-saver timeout in minutes (1 to 60), 0 disables pwr save"""
        return self._psave_holdoff

    @pwr_save.setter
    @validate
    def pwr_save(self, value):
        self._psave_holdoff = value
        if value:
            self._visa.write(f"DISPLAY:PSAVE:HOLDOFF {value}")
            self._visa.write("DISPLAY:PSAVE:STATE ON")
        else:
            self._visa.write("DISPLAY:PSAVE:STATE OFF")



"""Spectrum analyzer mode"""
import sys
import asyncio
import itertools
from dataclasses import dataclass
import numpy as np
from unyt import unyt_quantity, unyt_array
from rsfsup.common import Subsystem
from rsfsup.spectrum_analyzer.display import Display
from rsfsup.spectrum_analyzer.frequency import Frequency
from rsfsup.spectrum_analyzer.amplitude import Amplitude
from rsfsup.spectrum_analyzer.bandwidth import Bandwidth
from rsfsup.spectrum_analyzer.sweep import Sweep
from rsfsup.spectrum_analyzer.markers import Marker, DeltaMarker
from rsfsup.spectrum_analyzer.traces import Trace


@dataclass
class State:
    """Signal for controlling asyncio tasks"""

    running: bool = False


class SpecAn(Subsystem, kind="Spectrum Analyzer"):
    """Spectrum analyzer mode

    Attributes:
        instr (Fsup)
    """

    def __init__(self, instr):
        super().__init__(instr)
        self.display = Display(instr)
        self.frequency = Frequency(instr)
        self.amplitude = Amplitude(instr)
        self.bandwidth = Bandwidth(instr)
        self.sweep = Sweep(instr)
        self._markers = [Marker(instr, num) for num in range(1, 5)]
        self._deltamarkers = [DeltaMarker(instr, num) for num in range(1, 5)]
        self._traces = [Trace(instr, 1)]
        trace = self._traces[0]
        trace.state = "ON"
        setattr(self, trace._name, trace)
        self._state = State()

    def reference_fixed(self, state="ON"):
        """state (str): {ON, OFF}
            Turn on markers 1 and 2, if necessary, and set markers 2 to 4 as delta
            markers fixed to marker 1
        """
        self.enable_marker(1)
        self.enable_marker(2, as_delta=True)
        self._visa.write(f"CALC{self._screen()}:DELT2:FUNC:FIX {state}")

    def enable_marker(self, num, as_delta=False):
        """num (int): marker/deltamarker number {1, 2, 3, 4}
        as_delta (bool): True if deltamarker"""
        i = num - 1
        marker = self._markers[i]
        deltamarker = self._deltamarkers[i]
        # pylint: disable=protected-access
        if as_delta:
            deltamarker.state = "ON"
            try:
                setattr(self, deltamarker._name, deltamarker)
            except AttributeError:
                pass
            marker.state = "OFF"
            try:
                delattr(self, marker._name)
            except AttributeError:
                pass
        else:
            deltamarker.state = "OFF"
            try:
                delattr(self, deltamarker._name)
            except AttributeError:
                pass
            marker.state = "ON"
            try:
                setattr(self, marker._name, marker)
            except AttributeError:
                pass

    def turnoff_markers(self):
        """Turn off all markers in the active screen"""
        for marker in self._markers + self._deltamarkers:
            marker.state = "OFF"

    def clear_traces(self):
        """Clear all traces"""
        self._visa.write(f"DISP:WIND{self._screen()}:TRACE:CLEAR")

    async def _show_spinner(self):
        """Show an in-progress spinner during phase noise measurement"""
        spinner = itertools.cycle(["-", "/", "|", "\\"])
        try:
            while self._state.running:
                sys.stdout.write(next(spinner))
                sys.stdout.flush()
                sys.stdout.write("\b")
                await asyncio.sleep(0.5)
            return 0
        except asyncio.CancelledError:
            pass
        finally:
            sys.stdout.write("\b \b")

    async def _measure_spectrum(self):
        """Measure spectrum"""
        try:
            self._visa.write("*CLS")
            self._visa.write("SYSTEM:ERROR:CLEAR:ALL")
            self._visa.query("STATUS:OPERATION:EVENT?")
            self._visa.query("STATUS:QUESTIONABLE:EVENT?")
            self._visa.write("*SRE 191")
            self._visa.write("*ESE 63")
            self._visa.write("STATUS:OPERATION:PTRANSITION 0")
            self._visa.write("STATUS:QUESTIONABLE:PTRANSITION 296")
            original_continuous = self._visa.query("INIT:CONT?")
            self._visa.write("INIT:CONT OFF")
            self._visa.write(f"INIT; *OPC")
            # poll the ESB bit for an event occurance indicating completion or error
            while not self._visa.stb & 32:
                await asyncio.sleep(1)
            self._state.running = False
            esr = int(self._visa.query("*ESR?"))
            op_complete = bool(esr & 1)
            return 0 if op_complete else 1
        except asyncio.CancelledError:
            pass
        finally:
            self._visa.write(f"INIT:CONT {original_continuous}")

    async def _start_task(self, timeout):
        self._state.running = True
        task = asyncio.gather(self._show_spinner(), self._measure_spectrum())
        try:
            ret_value = await asyncio.wait_for(task, timeout)
        except asyncio.TimeoutError:
            raise TimeoutError("Phase noise measurement timed out") from None
        else:
            return ret_value

    def read(self, trace=1, timeout=None):
        """Read trace data and return tuple (X, Y)

        Parameters:
            trace (int): {1, 2, 3}
            timeout (int): timeout in seconds or None
        """
        ret_value = asyncio.run(self._start_task(timeout))
        if ret_value is None:
            return None
        if ret_value[1] > 0:
            print(self._instr.status.event_status)
            return None
        num = self.sweep.points
        start_str = self.frequency.start
        value, unit = start_str.split(" ")
        start = unyt_quantity(float(value), unit)
        stop_str = self.frequency.stop
        value, unit = stop_str.split(" ")
        stop = unyt_quantity(float(value), unit)
        x = np.linspace(start, stop, num=num, endpoint=True)
        x.name = "$f$"
        i = trace - 1
        data = self._traces[i].data
        unit = self._traces[i].y_unit
        y = unyt_array(data, unit)
        y.name = "$P$"
        return (x, y)

    def __dir__(self):
        for marker in self._markers + self._deltamarkers:
            if marker.state == "OFF":
                try:
                    delattr(self, marker._name)
                except AttributeError:
                    pass
        return super().__dir__()



"""Amplitude subsystem"""
from rsfsup.common import Subsystem, validate


class Amplitude(Subsystem, kind="AMPT"):
    """Amplitude subsystem

    Attributes:
        instr (Fsup)
    """

    def __init__(self, instr):
        super().__init__(instr)
        # For now, set the input attenuation to automatic to track the REF LEVEL.
        # This will automatically adjust the attenuation to keep the maximum signal
        # level (REF LEVEL) at the mixer to the default level of -25 dBm.
        self._visa.write("INPUT:ATTENUATION:AUTO ON")
        self._visa.write("INPUT:MIXER:AUTO ON")

    @property
    def coupling(self):
        """value (str): RF input coupling {AC, DC}"""
        return self._visa.query("INPUT:COUPLING?")

    @coupling.setter
    @validate
    def coupling(self, value):
        self._visa.write(f"INPUT:COUPLING {value}")

    @property
    def ref_level(self):
        """value (int or str): reference level as integer in unit dBm or 'x dBm'"""
        return self._visa.query(f"DISPLAY:WINDOW{self._screen()}:TRACE:Y:RLEVEL?")

    @ref_level.setter
    @validate
    def ref_level(self, value):
        self._visa.write(f"DISP:WIND{self._screen()}:TRAC:Y:RLEV {value}")



"""Marker subsystem"""
from rsfsup.common import Subsystem, validate, BOOLEAN, scale_frequency


class Marker(Subsystem, kind="Marker"):
    """A marker

    Attributes:
        instr (Fsup)
        num (int): marker number {1, 2, 3, 4}
    """

    def __init__(self, instr, num):
        super().__init__(instr)
        self._num = num
        self._name = f"marker_{num}"

    @property
    def state(self):
        """(str): {ON, OFF}"""
        result = self._visa.query(f"CALC{self._screen()}:MARK{self._num}:STATE?")
        return BOOLEAN.get(result, result)

    @state.setter
    @validate
    def state(self, value):
        self._visa.write(f"CALC{self._screen()}:MARK{self._num}:STATE {value}")

    @property
    def count(self):
        """value (str): frequency counter for marker 1 {ON, OFF}"""
        if self._num != 1:
            raise ValueError("Frequency counter only available on marker 1")
        result = self._visa.query(f"CALC{self._screen()}:MARK:COUNT?")
        return BOOLEAN.get(result, result)

    @count.setter
    @validate
    def count(self, value):
        if self._num != 1:
            raise ValueError("Frequency counter only available on marker 1")
        self._visa.write(f"CALC{self._screen()}:MARK:COUNT {value}")

    @property
    def frequency(self):
        """(str): measured frequency at marker 1"""
        value = float(self._visa.query(f"CALC{self._screen()}:MARK:COUNT:FREQUENCY?"))
        return scale_frequency(value)

    @property
    def resolution(self):
        """value (int or str): frequency counter resolution {0.1, 1, 10, ..., 10000} Hz"""
        value = float(self._visa.query(f"CALC{self._screen()}:MARK:COUNT:RESOLUTION?"))
        return scale_frequency(value)

    @resolution.setter
    @validate
    def resolution(self, value):
        self._visa.write(f"CALC{self._screen()}:MARK:COUNT:RESOLUTION {value}")

    @property
    def X(self):
        """value (int or str): 0 to f_max Hz"""
        value = float(self._visa.query(f"CALC{self._screen()}:MARK{self._num}:X?"))
        return scale_frequency(value)

    @X.setter
    @validate
    def X(self, value):
        self._visa.write(f"CALC{self._screen()}:MARK{self._num}:X {value}")

    @property
    def Y(self):
        """(str): level in dBm"""
        # change to single sweep, complete acquisition, measure power
        continuous = self._visa.query(f"INIT{self._screen()}:CONT?")
        original_continuous = BOOLEAN.get(continuous, continuous)
        state = self._visa.query(f"CALC{self._screen()}:MARK{self._num}:STATE?")
        original_state = BOOLEAN.get(state, state)
        original_timeout = self._visa.timeout
        sweep_time = float(self._visa.query(f"SENSE{self._screen()}:SWEEP:TIME?"))  # s
        self._visa.timeout = original_timeout + 1000 * sweep_time
        self._visa.write(f"INIT{self._screen()}:CONT OFF")
        self._visa.write(f"CALC{self._screen()}:MARK{self._num}:STATE ON")
        self._visa.write(f"INIT{self._screen()};*WAI")
        self._visa.timeout = original_timeout
        value = self._visa.query(f"CALC{self._screen()}:MARK{self._num}:Y?")
        self._visa.write(f"CALC{self._screen()}:MARK{self._num}:STATE {original_state}")
        self._visa.write(f"INIT{self._screen()}:CONT {original_continuous}")
        return f"{value} dBm"

    def move_to(self, location="PEAK"):
        """location (str): {PEAK, NEXT PEAK, NEXT PEAK RIGHT, NEXT PEAK LEFT}"""
        prefix = f"CALC{self._screen()}:MARK{self._num}:MAX"
        if location == "NEXT PEAK":
            suffix = "NEXT"
        elif location.endswith("RIGHT"):
            suffix = "RIGHT"
        elif location.endswith("LEFT"):
            suffix = "LEFT"
        else:
            suffix = "PEAK"
        self._visa.write(f"{prefix}:{suffix}")


class DeltaMarker(Subsystem, kind="Deltamarker"):
    """Deltamarker

    Attributes:
        instr (Fsup)
        num (int): deltamarker number {1, 2, 3, 4}
    """

    def __init__(self, instr, num):
        super().__init__(instr)
        self._num = num
        self._name = f"deltamarker_{num}"

    @property
    def state(self):
        """(str): {ON, OFF}"""
        result = self._visa.query(f"CALC{self._screen()}:DELT{self._num}?")
        return BOOLEAN.get(result, result)

    @state.setter
    @validate
    def state(self, value):
        self._visa.write(f"CALC{self._screen()}:DELT{self._num}:STATE {value}")

    @property
    def X(self):
        """value (int or str): 0 to f_max Hz"""
        value = float(self._visa.query(f"CALC{self._screen()}:DELT{self._num}:X?"))
        return scale_frequency(value)

    @property
    def Y(self):
        """(str): level in dBm"""
        # change to single sweep, complete acquisition, measure power
        continuous = self._visa.query(f"INIT{self._screen()}:CONT?")
        original_continuous = BOOLEAN.get(continuous, continuous)
        state = self._visa.query(f"CALC{self._screen()}:DELT{self._num}:STATE?")
        original_state = BOOLEAN.get(state, state)
        original_timeout = self._visa.timeout
        sweep_time = float(self._visa.query(f"SENSE{self._screen()}:SWEEP:TIME?"))  # s
        self._visa.timeout = original_timeout + 1000 * sweep_time
        self._visa.write(f"INIT{self._screen()}:CONT OFF")
        self._visa.write(f"CALC{self._screen()}:DELT{self._num}:STATE ON")
        self._visa.write(f"INIT{self._screen()};*WAI")
        self._visa.timeout = original_timeout
        value = self._visa.query(f"CALC{self._screen()}:DELT{self._num}:Y?")
        self._visa.write(f"CALC{self._screen()}:DELT{self._num}:STATE {original_state}")
        self._visa.write(f"INIT{self._screen()}:CONT {original_continuous}")
        return f"{value} dBm"

    def move_to(self, location="PEAK"):
        """location (str): {PEAK, NEXT PEAK, NEXT PEAK RIGHT, NEXT PEAK LEFT}"""
        prefix = f"CALC{self._screen()}:DELT{self._num}:MAX"
        if location == "NEXT PEAK":
            suffix = "NEXT"
        elif location.endswith("RIGHT"):
            suffix = "RIGHT"
        elif location.endswith("LEFT"):
            suffix = "LEFT"
        else:
            suffix = "PEAK"
        self._visa.write(f"{prefix}:{suffix}")



"""Status reporting system"""
from rsfsup.common import Subsystem

STB = [
    None,
    None,
    "Error queue not empty",
    "Questionable status register sum bit",
    "MAV, message available",
    "ESB, event status sum bit",
    "MSS, master status summary",
    "Operation status register sum bit",
]

ESR = [
    "Operation complete",
    None,
    "Query error",
    "Device-dependent error",
    "Execution error",
    "Command error",
    "User request (LOCAL)",
    "Power on",
]

OPER = [
    "Calibrating",
    None,
    None,
    None,
    "Measuring phase noise",
    None,
    None,
    None,
    "Hardcopy in progress",
    None,
    None,
    None,
    None,
    None,
    None,
    None,
]

QUES = [
    None,
    None,
    None,
    "RF power",
    None,
    "Frequency",
    None,
    "Low phase noise board",
    "Calibration",
    "Limit",
    "Margin",
    None,
    "Adjacent channel power",
    "Phase noise",
    "DC or phase noise board",
    None,
]

POW = [
    "RF input overload, screen A",
    "RF input underload, screen A",
    "IF path overload, screen A",
    "RF or IF overload occured during AVERAGE or MAX/MINHOLD, screen A",
    None,
    None,
    None,
    None,
    "RF input overload, screen B",
    "RF input underload, screen B",
    "IF path overload, screen B",
    "RF or IF overload occured during AVERAGE or MAX/MINHOLD, screen B",
    None,
    None,
    None,
    None,
]

LPN = 4 * [None] + ["PLL unlock"] + 11 * [None]

PNO = [
    "Invalid DC switching sequence",
    "Signal not found",
    "Verify signal failed",
    "Beat note not found",
    "Beat note level too low",
    "Reference frequency out of range",
    None,
    "Power to the DUT is disabled or disconnected",
    "No traces",
    "Automatic frequency control error",
] + 6 * [None]

DCPN = (
    [
        "Temperature of DC board too high",
        "Current Vcc1 too high",
        "Current Vcc2 too high",
        "Current Vaux too high",
        "Current Vtune1 too high",
        "Current Vtune2 too high",
        "Vtune1 above Vmax",
        "Vtune2 above Vmax",
    ]
    + 3 * [None]
    + ["PLL loop unlock"]
    + 4 * [None]
)


def decode_register(register, value):
    """Decode decimal value for register

    Attributes:
        register (list): list of bit descriptions
        value (int): 0 to 255

    Returns:
        list of descriptions of bits that are set
    """
    set_bits = []
    for bit_description in register:
        if value & 1:
            if bit_description is not None:
                set_bits.append(bit_description)
        value = value >> 1
    return set_bits


class Status(Subsystem, kind="Status"):
    """Status reporting system

    Attributes:
        instr (Fsup)
    """

    def clear(self):
        """Clear status reporting system (*CLS)"""
        self._visa.write("*CLS")
        self._visa.write("SYSTEM:ERROR:CLEAR:ALL")

    @property
    def status_byte(self):
        """Query Status Byte (*STB) register

        Returns tuple of integer value and descriptions for bits that are set
        """
        value = self._visa.stb
        set_bits = decode_register(STB, value)
        return (value, set_bits)

    @property
    def event_status(self):
        """Query Event Status register (*ESR)

        Returns tuple of integer value and descriptions for bits that are set
        """
        value = int(self._visa.query("*ESR?"))
        set_bits = decode_register(ESR, value)
        return (value, set_bits)

    @property
    def operation(self):
        """Query the Operation register - events

        Returns tuple of integer value and descriptions for bits that are set
        """
        value = int(self._visa.query("STATUS:OPERATION?"))
        set_bits = decode_register(OPER, value)
        return (value, set_bits)

    @property
    def questionable(self):
        """Query the Questionable register - events

        Returns tuple of integer value and descriptions for bits that are set
        """
        value = int(self._visa.query("STATUS:QUESTIONABLE?"))
        set_bits = decode_register(QUES, value)
        if value & 2 ** 3:
            sub_value = int(self._visa.query("STATUS:QUESTIONABLE:POWER?"))
            set_bits.append(decode_register(POW, sub_value))
        if value & 2 ** 7:
            sub_value = int(self._visa.query("STATUS:QUESTIONABLE:LPNOISE?"))
            set_bits.append(decode_register(LPN, sub_value))
        if value & 2 ** 13:
            sub_value = int(self._visa.query("STATUS:QUESTIONABLE:PNOISE?"))
            set_bits.append(decode_register(PNO, sub_value))
        if value & 2 ** 14:
            sub_value = int(self._visa.query("STATUS:QUESTIONABLE:DCPNOISE?"))
            set_bits.append(decode_register(DCPN, sub_value))
        return (value, set_bits)



"""Sweep subsystem"""
from rsfsup.common import Subsystem, validate, scale_time, BOOLEAN


class Sweep(Subsystem, kind="SWEEP"):
    """Sweep subsystem

    Attributes:
        instr (Fsup)
    """

    def __init__(self, instr):
        super().__init__(instr)
        # For now, adjust sweep time automatically according to span and bandwidth
        # settings.
        self._visa.write("SWEEP:TIME:AUTO ON")

    @property
    def time(self):
        """(str): 2.5 ms to 16000 s"""
        value = self._visa.query(f"SENSE{self._screen()}:SWEEP:TIME?")
        return scale_time(float(value))

    @property
    def continuous(self):
        """value (str): {ON, OFF}"""
        result = self._visa.query(f"INITIATE{self._screen()}:CONTINUOUS?")
        return BOOLEAN.get(result, result)

    @continuous.setter
    @validate
    def continuous(self, value):
        self._visa.write(f"INITIATE:{self._screen()}:CONTINUOUS {value}")

    @property
    def points(self):
        """value (int): {155, 313, 625, ..., 30001}"""
        return int(self._visa.query(f"SENSE{self._screen()}:SWEEP:POINTS?"))

    @points.setter
    @validate
    def points(self, value):
        self._visa.write(f"SENSE{self._screen()}:SWEEP:POINTS {value}")

    @property
    def count(self):
        """value (int): 0 to 32767"""
        return int(self._visa.query(f"SENSE{self._screen()}:SWEEP:COUNT?"))

    @count.setter
    @validate
    def count(self, value):
        self._visa.write(f"SENSE{self._screen()}:SWEEP:COUNT {value}")



"""Fsup class"""
from datetime import datetime
from rsfsup.common import RSBase, get_idn, validate
from rsfsup.status import Status
from rsfsup.spectrum_analyzer.instrument import SpecAn
from rsfsup.ssa.instrument import SSA


class Fsup(RSBase):
    """Fsup class

    Attributes:
        visa (pyvisa.resources.Resource): pyvisa resource
    """

    _modes = {"SAN": "SPECTRUM", "PNO": "SSA"}
    _options = {
        "B4": "Low-aging OCXO",
        "B10": "External Generator Control",
        "B18": "Removable Hard Disk",
        "B19": "Second Hard Disk",
        "B21": "LO/IF Ports for External Mixers",
        "B23": "20 dB Preamplifier",
        "B25": "Electronic Attenuator",
        "B28": "Trigger Port",
        "B60": "Low Phase Noise",
        "B61": "Correlation Extension",
        "K5": "GSM/EDGE",
        "K8": "Bluetooth",
        "K9": "Power Sensor Measurements",
        "K30": "Noise Figure and Gain Measurements",
        "K70": "Vector Signal Analysis",
    }

    def __init__(self, visa):
        super().__init__(visa)
        self._visa.write("*CLS")
        self._visa.write("SYSTEM:ERROR:CLEAR:ALL")
        self._visa.write("*ESE 255")
        self._visa.write("FORMAT ASCII")
        self._idn = get_idn(visa)
        self._currentmode = self.mode
        if self._currentmode == "SPECTRUM":
            self.spectrum = SpecAn(self)
        else:
            self.ssa = SSA(self)
        self.status = Status(self)

    @property
    def model(self):
        """(str): the model number"""
        return self._idn.model

    @property
    def serial_number(self):
        """(str): the serial number"""
        return self._idn.serial_number

    @property
    def firmware_version(self):
        """(str): the firmware version"""
        return self._idn.firmware_version

    @property
    def mode(self):
        """value (str): {SPECTRUM, SSA}"""
        result = self._visa.query("INSTRUMENT?")
        return Fsup._modes.get(result, result)

    @mode.setter
    @validate
    def mode(self, value):
        modes = dict((v, k) for k, v in Fsup._modes.items())
        if value in modes and value != self._currentmode:
            original_timeout = self._visa.timeout
            self._visa.timeout = 5000  # ms
            if self._currentmode == "SPECTRUM":
                del self.spectrum
            else:
                del self.ssa
            self._currentmode = value
            new_mode = modes[value]
            self._visa.write(f"INSTRUMENT:SELECT {new_mode}")
            if self._currentmode == "SPECTRUM":
                self.spectrum = SpecAn(self)
            else:
                self.ssa = SSA(self)
            self._visa.timeout = original_timeout

    def reset(self):
        """Reset the instrument to the factory default settings"""
        self._visa.write("*RST")

    def set_clock(self):
        """Set the date and time to local machine's time"""
        self._visa.write(f"SYSTEM:DATE {datetime.now().strftime('%Y,%m,%d')}")
        self._visa.write(f"SYSTEM:TIME {datetime.now().strftime('%H,%M,%S')}")

    def lock_frontpanel(self):
        """Lock the front panel"""
        self._visa.write("SYSTEM:DISPLAY:FPANEL OFF")
        self._visa.write("SYSTEM:KLOCK ON")

    def unlock_frontpanel(self):
        """Unlock the front panel"""
        self._visa.write("SYSTEM:KLOCK OFF")
        self._visa.write("SYSTEM:DISPLAY:FPANEL ON")

    @property
    def options(self):
        """List the installed options"""
        codes = self._visa.query("*OPT?").split(",")
        installed_codes = [code for code in codes if not code.startswith("0")]
        return [Fsup._options.get(code, code) for code in installed_codes]

    @property
    def impedance(self):
        """value (int): RF input impedance {50, 75} Ω"""
        return int(self._visa.query("INPUT:IMPEDANCE?"))

    @impedance.setter
    @validate
    def impedance(self, value):
        self._visa.write(f"INPUT:IMPEDANCE {value}")

    def __repr__(self):
        return f"<R&S {self.model} at {self._visa.resource_name}>"



"""Phase noise measurement configuration"""
from rsfsup.common import Subsystem, validate, scale_frequency


class Configure(Subsystem, kind="Phase Noise Configuration"):
    """Phase noise configuration

    Constraints:
        Automatic preamplifier and attenuator control turned ON
        Automatic Frequency Control (AFC) turned OFF
        Automatic frequency and level verification turned ON
        Optimize cross-correlation if option present
        Automatic configuration of the internal VCO & PLL settings
        Suppress all spurs

    Attributes:
        instr (Fsup)
    """

    _testsetups = {
        "INT1": "Default PLL mode using internal generator, phase detector and control",
        "INT3": "Cross-correlation using internal generator, phase detector and control",
    }
    _search_bands = [
        "1 MHz to 10 MHz",
        "10 MHz to 50 MHz",
        "50 MHz to 8 GHz",
        "8 GHz to 26.5 GHz",
        "26.5 GHz to 50 GHz",
    ]

    def __init__(self, instr):
        super().__init__(instr)
        self._visa.write("INPUT:GAIN:AUTO ON")
        self._visa.write("FREQUENCY:TRACK OFF")
        self._visa.write("FREQUENCY:VERIFY:STATE ON")
        self._visa.write("SWEEP:XOPTIMIZE ON")
        self._visa.write("VCO:CSEARCH:AUTO ON")
        self._visa.write("VCO:LOOP:BWIDTH:AUTO ON")
        self._visa.write("VCO:PDETECTOR:PREMEAS:STATE ON")
        self._visa.write("VCO:PREMEAS:STATE ON")
        self._visa.write("SPURS:HIGHLIGHT:STATE OFF")
        self._visa.write("SPURS:SUPPRESS ALL")

    @property
    def test_setup(self):
        """value (str): {INT1, INT3}"""
        ts = self._visa.query("CONFIGURE:TSETUP?")
        return f"{ts}: {Configure._testsetups[ts]}"

    @test_setup.setter
    @validate
    def test_setup(self, value):
        self._visa.write(f"CONFIGURE:TSETUP {value}")

    @property
    def rf_level_detection(self):
        """value (str): {AUTO, MANUAL}"""
        auto_result = int(self._visa.query("CONFIGURE:POWER:AUTO?"))
        if auto_result:
            detection_mode = "AUTO"
        else:
            detection_mode = "MANUAL"
        return detection_mode

    @rf_level_detection.setter
    @validate
    def rf_level_detection(self, value):
        if value == "MANUAL":
            self._visa.write("CONFIGURE:POWER:AUTO OFF")
        else:
            self._visa.write("CONFIGURE:POWER:AUTO ON")

    @property
    def expected_rf_level(self):
        """value (float, str): -100 dBm to 30 dBm"""
        return float(self._visa.query("CONFIGURE:POWER:EXPECTED:RF?"))

    @expected_rf_level.setter
    @validate
    def expected_rf_level(self, value):
        self._visa.write(f"CONFIGURE:POWER:EXPECTED:RF {value}")

    @property
    def start_frequency(self):
        """value (int, str): start frequency in hertz in discrete 1, 3 steps"""
        value = int(self._visa.query("FREQUENCY:START?"))
        return scale_frequency(value)

    @start_frequency.setter
    @validate
    def start_frequency(self, value):
        self._visa.write(f"FREQUENCY:START {value}")

    @property
    def stop_frequency(self):
        """value (int, str): stop frequency in hertz in discrete 1, 3 steps"""
        value = int(self._visa.query("FREQUENCY:STOP?"))
        return scale_frequency(value)

    @stop_frequency.setter
    @validate
    def stop_frequency(self, value):
        self._visa.write(f"FREQUENCY:STOP {value}")

    @property
    def yaxis_autoscale(self):
        """value (str): {ON, OFF, ONCE}"""
        return self._visa.query("DISPLAY:TRACE:Y:SCALE:AUTO?")

    @yaxis_autoscale.setter
    @validate
    def yaxis_autoscale(self, value):
        self._visa.write(f"DISPLAY:TRACE:Y:SCALE:AUTO {value}")

    @property
    def yaxis_top(self):
        """value (int, str): y-axis top in dBc/Hz"""
        return int(self._visa.query("DISPLAY:TRACE:Y:SCALE:RLEVEL?"))

    @yaxis_top.setter
    @validate
    def yaxis_top(self, value):
        self._visa.write(f"DISPLAY:TRACE:Y:SCALE:RLEVEL {value}")

    @property
    def yaxis_range(self):
        """value (int, str): y-axis top in dBc/Hz"""
        return int(self._visa.query("DISPLAY:TRACE:Y:SCALE?"))

    @yaxis_range.setter
    @validate
    def yaxis_range(self, value):
        self._visa.write(f"DISPLAY:TRACE:Y:SCALE {value}")

    @property
    def sweep_mode(self):
        """value (str): {FAST, NORMAL, AVERAGED}"""
        return self._visa.query("SWEEP:MODE?")

    @sweep_mode.setter
    @validate
    def sweep_mode(self, value):
        self._visa.write(f"SWEEP:MODE {value}")

    @property
    def sweep_count(self):
        """value (int): number of measurement sweeps"""
        return int(self._visa.query("SWEEP:COUNT?"))

    @sweep_count.setter
    @validate
    def sweep_count(self, value):
        self._visa.write(f"SWEEP:COUNT {value}")

    @property
    def measurement_time(self):
        """(float): estimated total measurement time in seconds"""
        return float(self._visa.query("SWEEP:TIME?"))

    @property
    def dut_type(self):
        """value (str): {AUTO, FRLP, SYNT, OCXO, XTAL}"""
        return self._visa.query("VCO:TYPE?")

    @dut_type.setter
    @validate
    def dut_type(self, value):
        self._visa.write(f"VCO:TYPE {value}")

    @property
    def frequency_search_bands(self):
        """value (int, str): {ALL, 2 to 4}"""
        bands = []
        for band in range(2, 5):
            state = int(self._visa.query(f"VCO:BAND{band}?"))
            if state:
                bands.append((band, Configure._search_bands[band - 1]))
        return bands

    @frequency_search_bands.setter
    @validate
    def frequency_search_bands(self, value):
        for band in range(2, 5):
            if value in (band, "ALL"):
                self._visa.write(f"VCO:BAND{band} ON")
            else:
                self._visa.write(f"VCO:BAND{band} OFF")



"""Bandwidth subsystem"""
from rsfsup.common import Subsystem, scale_frequency


class Bandwidth(Subsystem, kind="BW"):
    """Bandwidth subsystem

    Attributes:
        instr (Fsup)
    """

    def __init__(self, instr):
        super().__init__(instr)
        # For now, adjust resolution and video bandwidths automatically according to
        # the span.
        self._visa.write("BANDWIDTH:RESOLUTION:AUTO ON")
        self._visa.write("BANDWIDTH:RESOLUTION:TYPE NORMAL")
        self._visa.write("BANDWIDTH:VIDEO:AUTO ON")

    @property
    def resolution_bandwidth(self):
        """(str): 10 Hz to 50 MHz"""
        value = self._visa.query(f"SENSE{self._screen()}:BANDWIDTH:RESOLUTION?")
        return scale_frequency(float(value))

    @property
    def video_bandwidth(self):
        """(str): 1 Hz to 10 MHz"""
        value = self._visa.query(f"SENSE{self._screen()}:BANDWIDTH:VIDEO?")
        return scale_frequency(float(value))



"""Setuptools install script"""
from setuptools import setup, find_packages


def get_readme():
    """Get readme contents"""
    with open("README.md") as f:
        return f.read()


def get_version():
    """Get the version number"""
    with open("rsfsup/version.py") as f:
        lines = f.readlines()
    line = ""
    for line in lines:
        if line.startswith("__version__"):
            break
    version = [s.strip().strip('"') for s in line.split("=")][1]
    return version


setup(
    name="rsfsup",
    version=get_version(),
    author="Lee Johnston",
    author_email="lee.johnston.100@gmail.com",
    description="Communication with Rohde&Schwarz FSUP Signal Source Analyzer",
    long_description=get_readme(),
    long_description_content_type="text/markdown",
    packages=find_packages(),
    license="MIT",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Science/Research",
    ],
    url="https://github.com/l-johnston/rsfsup",
    install_requires=["pyvisa", "unyt"],
)



"""Signal Source Analyzer mode (PNO)"""
import sys
import itertools
import asyncio
from dataclasses import dataclass
from unyt import unyt_array
from rsfsup.common import Subsystem
from rsfsup.ssa.display import Display
from rsfsup.ssa.configure import Configure
from rsfsup.ssa.traces import Trace


@dataclass
class State:
    """Signal for controlling asyncio tasks"""

    running: bool = False


class SSA(Subsystem, kind="SSA"):
    """Signal Source Analyzer subsystem for phase noise measurements

    Steps to measure phase noise of a CW signal source (PLL, CCORrelation):
    - Set center frequency, amplitude and bandwidth in SPECTRUM mode
    - Switch to SSA mode
    - Configure measurement (PLL, CCORrelation)
    - Configure expected RF power ('auto', float dBm)
    - Configure TSETup {INT1, INT3}

    Attributes:
        instr (Fsup)
    """

    def __init__(self, instr):
        super().__init__(instr)
        self.display = Display(instr)
        self.configure = Configure(instr)
        self._traces = [Trace(instr, 1)]
        trace = self._traces[0]
        trace.state = "ON"
        setattr(self, trace._name, trace)
        self._state = State()

    async def _show_spinner(self):
        """Show an in-progress spinner during phase noise measurement"""
        spinner = itertools.cycle(["-", "/", "|", "\\"])
        try:
            while self._state.running:
                sys.stdout.write(next(spinner))
                sys.stdout.flush()
                sys.stdout.write("\b")
                await asyncio.sleep(0.5)
            return 0
        except asyncio.CancelledError:
            pass
        finally:
            sys.stdout.write("\b \b")

    async def _measure_phase_noise(self, premeasure):
        """Perform phase noise measurement"""
        sweep = "NEW" if premeasure else "IMMEDIATE"
        try:
            self._visa.write("*CLS")
            self._visa.write("SYSTEM:ERROR:CLEAR:ALL")
            self._visa.query("STATUS:OPERATION:EVENT?")
            self._visa.query("STATUS:QUESTIONABLE:EVENT?")
            self._visa.write("*SRE 191")
            self._visa.write("*ESE 63")
            self._visa.write("STATUS:OPERATION:PTRANSITION 16")
            self._visa.write("STATUS:QUESTIONABLE:PTRANSITION 24744")
            original_continuous = self._visa.query("INIT:CONT?")
            self._visa.write("INIT:CONT OFF")
            self._visa.write(f"INIT:{sweep}; *OPC")
            # poll the ESB bit for an event occurance indicating completion or error
            while not self._visa.stb & 32:
                await asyncio.sleep(1)
            self._state.running = False
            esr = int(self._visa.query("*ESR?"))
            op_complete = bool(esr & 1)
            return 0 if op_complete else 1
        except asyncio.CancelledError:
            pass
        finally:
            self._visa.write(f"INIT:CONT {original_continuous}")

    async def _start_task(self, premeasure, timeout):
        self._state.running = True
        task = asyncio.gather(
            self._show_spinner(), self._measure_phase_noise(premeasure)
        )
        try:
            ret_value = await asyncio.wait_for(task, timeout)
        except asyncio.TimeoutError:
            raise TimeoutError("Phase noise measurement timed out") from None
        else:
            return ret_value

    def read(self, premeasure=True, timeout=None):
        """Measure the phase noise and return (X, Y) tuple

        Parameters:
            premeasure (bool): perform premeasurement
            timeout (int): timeout in seconds or None
        """
        ret_value = asyncio.run(self._start_task(premeasure, timeout))
        if ret_value is None:
            return None
        if ret_value[1] > 0:
            print(self._instr.status.event_status)
            return None
        x_values, y_values = self._traces[0].data
        x = unyt_array(x_values, "Hz")
        x.name = "$f$"
        y = unyt_array(y_values, "dBc/Hz")
        y.name = "$N$"
        return (x, y)
